/**************************************************************************\
|                                ECHAT v1.01                               |
|                         freeware by David Phillips                       |
|                                April 1998                                |
|                                                                          |
| For use with Watcom under DOS/4GW.  You are free to use this code in any |
| way you see fit, so long as you email me at electrum@tfs.net.  I would   |
| to know who is using my work.  Credit would be nice, but I do not expect |
| any, since this is a compilation of everyone else's work.  But isn't     |
| that really what programming is?                                         |
|                                                                          |
| The IPX routines were taken from ipxtest from abusesrc.zip, available    |
| from x2ftp.oulu.fi.  This is also where doscall came from.  The basic    |
| idea for this chat program was taken from ipxpmode.zip, also at x2ftp,   |
| though the entire chat portion was written from scratch by me.  I have   |
| released it because ipxpmode did not suit my purposes, and it was the    |
| only example of IPX I could find for Watcom.  I do not want to offend    |
| the author of ipxpmode, but provide another example.  I also plan to     |
| release Watcom Red Book Audio, available where you obtained this file.   |
|                                                                          |
| If you would like to work on a project with me, please email.            |
\**************************************************************************/



#include "doscall.cpp"
#include <dos.h>
#include <string.h>
#include <graph.h>
#include <ctype.h>
#include <conio.h>
#include <i86.h>

#define JC_SEG(x) (((long)x)>>4)
#define JC_OFF(x) (((long)x)&0xf)
#define MKPTR(x)  ((   (((unsigned long)x)&0xffff0000)>>12)+  (((unsigned long)x)&0xffff))
#define MAX_PACKETS 10
#define jmalloc(x,str) malloc(x)
#define jfree(x)       free(x)

#pragma pack (1)

static rminfo rm;
#include <stdlib.h>

struct ECBStructure
{
	ushort Link[2];                /* offset-segment */
	ushort ESRAddress[2];          /* offset-segment */
	uchar  InUseFlag;
	uchar  CompletionCode;
	ushort ECBSocket;              /* high-low */
	uchar  IPXWorkspace[4];        /* N/A */
	uchar  DriverWorkspace[12];    /* N/A */
	uchar  ImmediateAddress[6];    /* high-low */
	ushort FragmentCount;          /* low-high */
	ushort fAddress[2];            /* offset-segment */
	ushort fSize;                  /* low-high */
} ;

struct IPXPacketStructure
{
	ushort PacketCheckSum;         /* high-low */
	ushort PacketLength;           /* high-low */
	uchar  PacketTransportControl;
	uchar  PacketType;

	uchar  dNetwork[4];            /* high-low */
	uchar  dNode[6];               /* high-low */
	ushort dSocket;                /* high-low */

	uchar  sNetwork[4];            /* high-low */
	uchar  sNode[6];               /* high-low */
	ushort sSocket;                /* high-low */
};


struct JC_ipx_packet
{
	ECBStructure ecb;
	IPXPacketStructure ipx;
	ulong time_stamp;
	ushort verify_stamp;       // should be 0cdc  "crack dot com", all others ignored
	char buffer[512];          // DP -- was uchar, changed for strcpy()
} ;


int ipx_init()
{
	memset(&rm,0,sizeof(rm));         // clear real-mode (rm) interrupt regs
	rm.eax=0x7a00;                    // call ipx int
	RM_intr(0x2f,&rm);                // real-mode interrupt (rm int)
	return (rm.eax&0xff)==0xff;       // it should equal 0xff (returns 0 or 1)
}



uchar *ipx_get_local_address()                // same format as above (be sure to jfree this)
{
	uchar *addr=(uchar *)jmalloc(10,"IPX address");  // normal memory return
	uchar *low_addr=(uchar *)alloc_low_memory(10);   // low memory
	memset(&rm,0,sizeof(rm));                        // clear rm int regs
	rm.ebx=9;                                        // ???
	rm.esi=JC_OFF(low_addr);                         // offset of low mem
	rm.es=(short)JC_SEG(low_addr); // DP             // segment of low mem
	RM_intr(0x7a,&rm);                               // ipx int

	memcpy(addr,low_addr,10);       // copy from low mem to pmode mem
	free_low_memory(low_addr);      // free the temp low memory
	return addr;                    // return the address (pmode mem)
}


int ipx_listen(ECBStructure *ecb)
{
	memset(&rm,0,sizeof(rm));         // clear rm int regs
	rm.esi=JC_OFF(ecb);               // offset of ECBStructure
	rm.es=(short)JC_SEG(ecb); // DP   // segment of ecb

	rm.ebx=4;                         // ???
	RM_intr(0x7a,&rm);                // ipx int

	if (rm.eax&0xff)                  // 0xff equals failure
		return -1;                      // return failure
	return 0;                         // must be success
}

#define htons(x) ((((x)>>8)&0xff)|  (((x)&0xff)<<8))      // ???

JC_ipx_packet *ipx_socket(int port)
{
	memset(&rm,0,sizeof(rm));      // clear rm int regs
	rm.ebx=0;                      // Open socket function
	rm.eax=0;                      // close on program exit
	rm.edx=htons(port);            // dynamically allocate socket number
	RM_intr(0x7a,&rm);             // ipx int
	if (rm.eax&0xff)               // 0xff equals failure to open socket
	{
		printf("Unable to created ipx socket on port %d\n",port);
		return NULL;
	}
	//printf("opened socket %x\n",htons(port));       // print message

	//printf("created socket %x\n",rm.edx&0xffff);    // another message
	int fd=(ushort)htons((rm.edx&0xffff));          // ???  (socket we're on?)


	// allocate low memory for all of the packets
	JC_ipx_packet *pk=(JC_ipx_packet *)alloc_low_memory(sizeof(JC_ipx_packet)*MAX_PACKETS);
	if (!pk)
	{
		printf("unable to allocate low memory for packets\n");
		exit(0);
	}
	memset(pk,0,sizeof(JC_ipx_packet)*MAX_PACKETS);   // clear out all packets


	// setup an outgoing packet structure
	pk[0].ecb.ECBSocket = (short)htons(fd); // DP   // ???
	pk[0].ecb.FragmentCount = 1;                    // ???
	pk[0].ecb.fAddress[0] = (short)JC_OFF(&pk[0].ipx);  // DP -- addr offset
	pk[0].ecb.fAddress[1] = (short)JC_SEG(&pk[0].ipx);  // DP -- addr segment

//  pk[0].ipx.PacketCheckSum=0xffff;
 
	uchar *my_addr=ipx_get_local_address();       // get this comp's address
	memcpy(pk[0].ipx.sNetwork,my_addr,10);        // copy address to packet
	pk[0].ipx.sSocket=(short)htons(fd); // DP     // ???

	memcpy(pk[0].ipx.dNetwork,my_addr,4);         // copy part of address ???
	memset(pk[0].ipx.dNode,0xff,6);               // fill with 0xff (broadcast)
	memset(pk[0].ecb.ImmediateAddress,0xff,6);    // fill with 0xff (broadcast)
	pk[0].ipx.dSocket=(short)htons(fd); // DP     // ???

	jfree(my_addr);                               // free tempory memory

	
	for (int i=1;i<MAX_PACKETS;i++)               // run through all packets
	{
		// setup incoming packet structure
		pk[i].ecb.InUseFlag = 0x1d;                   // in-use flag == 0x1d
		pk[i].ecb.ECBSocket = (short)htons(fd); // DP // ???
		pk[i].ecb.FragmentCount = 1;                  // ???
		pk[i].ecb.fAddress[0] = (short)JC_OFF(&pk[i].ipx);  // DP -- it's offset
		pk[i].ecb.fAddress[1] = (short)JC_SEG(&pk[i].ipx);  // DP -- it's segment
		pk[i].ecb.fSize = sizeof(JC_ipx_packet)-sizeof(ECBStructure); // pkt size
		if (ipx_listen(&pk[i].ecb)<0)                 // ???
			printf("listen failed on packet %d\n",i);
		else
			;//printf("listen success on packet %d\n",i);
	}  

	//printf("listen to socket %x\n",htons(fd));    // ??? (socket we're on)
	return pk;                                    // return pointer to packets
}

int ipx_ready_to_read(JC_ipx_packet *pk)
{
	int i;
	for (i=1 ; i<MAX_PACKETS ; i++)             // run through all packets
	{
		if (!pk[i].ecb.InUseFlag)                 // if it's not in use
		{
			//printf("receive : <saw a packet>\n");   // message
			//printf("receive : %s\n", pk[i].buffer);
			if (ipx_listen(&pk[i].ecb)<0)           // if listen failed
			{ printf("failed to re-listen\n"); }    // message
			return i;                               // return which packet
		}
	}
	return 0;                                   // must be a failure
}

void ipx_close(JC_ipx_packet *pk)
{
	memset(&rm,0,sizeof(rm));                         // clear rm int regs
	rm.ebx=1;                                         // ???
	rm.edx=pk[0].ecb.ECBSocket;                       // ??? (socket we're on)
	//printf("close socket %x\n",pk[0].ecb.ECBSocket);  // message
	RM_intr(0x7a,&rm);                                // ipx int
}


void free_up_memory() { ; }


int ipx_send(JC_ipx_packet *pk)
{
	pk[0].ecb.fSize = sizeof(pk[0].ipx) + 512;    // ???

	unsigned char *a=(unsigned char *)pk[0].ipx.dNetwork;  // ptr to node/addr
	//printf("send : %x%x%x%x-%x%x%x%x%x%x (%x)\n",a[0],a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9],pk[0].ipx.dSocket);
	//printf("send : %s\n", pk[0].buffer);  // confirm data

	memset(&rm,0,sizeof(rm));           // clear rm int regs
	rm.ebx=3;                           // send packet function
	rm.esi=JC_OFF(&pk[0]);              // offset of packet
	rm.es=(short)JC_SEG(&pk[0]); // DP  // segment of packet

	RM_intr(0x7a,&rm);                  // ipx int
	return 1;                           // return 1
}

char *ipx_receive(JC_ipx_packet *pk)
{
	int receiver;

	receiver = ipx_ready_to_read(pk);
	if (receiver != 0)
		return pk[receiver].buffer;
	else
		return (char *)0;
}

void strtolower(char *string)
{
	int i;

	for(i = 0; i < strlen(string); i++)
		string[i] = (char)tolower(string[i]);
}

void print(char *what)
{
	char buf[5];
	int  x;

	_settextcolor(9);
	for(x = 0; x < strlen(what); x++)
	{
		if (what[x] == ' ')
			_settextcolor(2);
		sprintf(buf, "%c", what[x]);
		_outtext(buf);
	}
}

void main()
{
	JC_ipx_packet *sock;

	if (!ipx_init())
	{
		printf("IPX not installed\n");
		return;
	}
	else
	{
		sock = ipx_socket(0x869b);       // open socket
		if (sock==NULL)
		{
			printf("unable to open socket\n");
			return;
		}
	}
				
	// ipx is online!
	char name[255],
			 nameonly[255],
			 buffer[255],
			 buf[5],
			 *sent,
			 getc;
	int  len,
			 line,
			 size,
			 x;
	int  command,
			 beep;

	printf("\nECHAT v1.01 -- Freeware (c) 1998 by David Phillips\n");
	
	printf("name: ");
	scanf("%s", name);
	strcpy(nameonly, name);
	strcat(name, "> ");
	len = strlen(name);

	strcpy(sock[0].buffer, nameonly);
	strcat(sock[0].buffer, " has entered the chat room!");
	ipx_send(sock);

	strcpy(buffer, name);
	size = len;
	if (len > 20)
	{
		ipx_close(sock);
		printf("\nname is > 20 chars\n");
		return;
	}

	_setvideomode(_TEXTC80);
	_settextcolor(11);
	_setbkcolor(1);
	_settextposition(1, 17);
	_outtext("ECHAT v1.01 -- Freeware (c) 1998 by David Phillips\n");
	line = 2;
	_settextposition(25, 1);
	for(x = 0; x < 79; x++)
		_outtext(" ");
	_settextposition(25, 1);
	_outtext(buffer);
	_settextposition(25, (short)(size + 1));

	command = 0;
	beep = 0;

	do
	{
		if (kbhit())
			getc = ((char)getch());
		else
			getc = 0;
		if (getc == 9)
		{
			memset(buffer, 0, 255);
			strcpy(buffer, "COMMAND--> ");
			len = strlen("COMMAND--> ");
			size = len;
			_settextcolor(11);
			_setbkcolor(1);
			_settextposition(25, 1);
			for(x = 0; x < 79; x++)
				_outtext(" ");
			_settextposition(25, 1);
			_outtext(buffer);
			command = 1;
		}
		if (isprint(getc))
		{
			buffer[size] = getc;
			size++;
			_settextcolor(11);
			_setbkcolor(1);
			_settextposition(25, (short)size);
			sprintf(buf, "%c", getc);
			_outtext(buf);
		}
		if (((getc == 13) && (size > len)) || (size > 75))
		{
			if (command == 0)
			{
				strcpy(sock[0].buffer, buffer);
				ipx_send(sock);
			}
			else
			{
				strtolower(buffer);
				if ((strcmp(&buffer[11], "clear") == 0)
						|| (strcmp(&buffer[11], "cls") == 0))
				{
					for(x = 0; x < 24; x++)
					{
						_setbkcolor(0);
						_settextwindow(2, 1, 24, 80);
						_scrolltextwindow(1);
						_settextwindow(2, 1, 25, 80);
						line = 1;
					}
				}
				if ((strcmp(&buffer[11], "beep") == 0))
				{
					if (beep == 0)
						beep = 1;
					else
						beep = 0;
				}

			}
			len = strlen(name);
			memset(buffer, 0, 255);
			strcpy(buffer, name);
			size = len;
			_settextcolor(11);
			_setbkcolor(1);
			_settextposition(25, 1);
			for(x = 0; x < 79; x++)
				_outtext(" ");
			_settextposition(25, 1);
			_outtext(buffer);
			command = 0;
		}
		if ((getc == 8) && (size > len))
		{
			size--;
			buffer[size] = ' ';
			_settextposition(25, (short)(size + 1));
			_outtext(" ");
			_settextposition(25, (short)(size + 1));
		}
		sent = ipx_receive(sock);
		if (sent != 0)
		{
			if (beep == 1)
			{
				sound(800);
				delay(200);
				nosound();
			}
			_settextcolor(2);
			_setbkcolor(0);
			if (line < 24)
			{
				_settextposition((short)line, 1);
				print(sent);
				line++;
			}
			else
			{
				_settextposition(23, 1);
				print(sent);
				_setbkcolor(0);
				_settextwindow(2, 1, 24, 80);
				_scrolltextwindow(1);
				_settextwindow(2, 1, 25, 80);
			}
			_settextposition(25, (short)(size + 1));
		}
	} while (getc != 27);

	strcpy(sock[0].buffer, nameonly);
	strcat(sock[0].buffer, " has left the chat room!");
	ipx_send(sock);

	_setbkcolor(0);
	_settextcolor(7);
	_setvideomode(_DEFAULTMODE);
					 
	ipx_close(sock);
}
